# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for grouper."""

import datetime
import unittest

from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import grouper


@test_utils.with_cloud_emulators('datastore')
class GrouperTest(unittest.TestCase):
  """Grouper tests."""

  def setUp(self):
    self.testcases = [
        test_utils.create_generic_testcase(),
        test_utils.create_generic_testcase()
    ]

  def test_same_crash_different_security(self):
    """Test that crashes with same crash states, but different security
      flags."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_state = 'abc\ndef'
    self.testcases[1].security_flag = True
    self.testcases[1].crash_state = 'abc\ndef'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_same_crash_same_security(self):
    """Test that crashes with same crash states and same security flags get
    de-duplicated with one of them removed."""
    for index, t in enumerate(self.testcases):
      t.security_flag = True
      t.crash_state = 'abc\ndef'
      t.timestamp = datetime.datetime.utcfromtimestamp(index)
      t.put()

    grouper.group_testcases()

    testcases = []
    for testcase_id in data_handler.get_open_testcase_id_iterator():
      testcases.append(data_handler.get_testcase_by_id(testcase_id))

    self.assertEqual(len(testcases), 1)
    self.assertEqual(testcases[0].group_id, 0)
    self.assertTrue(testcases[0].is_leader)

  def test_unminimized(self):
    """Test that unminimized testcase is not processed for grouping."""
    self.testcases[0].security_flag = True
    self.testcases[0].crash_state = 'abc\ndef'
    self.testcases[0].crash_type = 'Heap-buffer-overflow\nREAD {*}'
    self.testcases[0].minimized_keys = None
    self.testcases[1].security_flag = True
    self.testcases[1].crash_state = 'abc\ndef'
    self.testcases[1].crash_type = 'Heap-buffer-overflow\nREAD 3'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    testcases = []
    for testcase_id in data_handler.get_open_testcase_id_iterator():
      testcases.append(data_handler.get_testcase_by_id(testcase_id))

    self.assertEqual(len(testcases), 2)
    self.assertEqual(testcases[0].group_id, 0)
    self.assertFalse(testcases[0].is_leader)
    self.assertEqual(testcases[1].group_id, 0)
    self.assertTrue(testcases[1].is_leader)

  def test_different_crash_same_security(self):
    """Test that crashes with different crash states and same security flags
      don't get grouped together."""
    self.testcases[0].security_flag = True
    self.testcases[0].crash_state = 'abc\ndef'
    self.testcases[1].security_flag = True
    self.testcases[1].crash_state = 'uvw\nxyz'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_group_of_one(self):
    """Test that a group id with just one testcase gets removed."""
    self.testcases[0].group_id = 1
    self.testcases[0].put()
    self.testcases[1].key.delete()

    grouper.group_testcases()

    testcase = data_handler.get_testcase_by_id(self.testcases[0].key.id())
    self.assertEqual(testcase.group_id, 0)
    self.assertTrue(testcase.is_leader)

  def test_same_unique_crash_type_with_same_state(self):
    """Test that the crashes with same unique crash type and same state get
    de-duplicated with one of them removed.."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Timeout'
    self.testcases[0].crash_state = 'abcde'
    self.testcases[0].timestamp = datetime.datetime.utcfromtimestamp(0)
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'Timeout'
    self.testcases[1].crash_state = 'abcde'
    self.testcases[1].timestamp = datetime.datetime.utcfromtimestamp(1)

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    testcases = []
    for testcase_id in data_handler.get_open_testcase_id_iterator():
      testcases.append(data_handler.get_testcase_by_id(testcase_id))

    self.assertEqual(len(testcases), 1)
    self.assertEqual(testcases[0].group_id, 0)
    self.assertTrue(testcases[0].is_leader)

  def test_same_unique_crash_type_with_different_state(self):
    """Test that the crashes with same unique crash type but different state
    don't get grouped together."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Timeout'
    self.testcases[0].crash_state = 'abcdef'
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'Timeout'
    self.testcases[1].crash_state = 'abcde'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_different_unique_crash_type_with_same_state(self):
    """Test that the crashes with different unique crash type but same state
    don't get grouped together."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Timeout'
    self.testcases[0].crash_state = 'abcde'
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'Out-of-memory'
    self.testcases[1].crash_state = 'abcde'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_different_unique_crash_type_with_different_state(self):
    """Test that the crashes with different unique crash type and different
    state don't get grouped together."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Timeout'
    self.testcases[0].crash_state = 'abcdef'
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'Out-of-memory'
    self.testcases[1].crash_state = 'abcde'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_different_crash_type_with_similar_state(self):
    """Test that the crashes with different crash types (one of them unique) and
    similar crash state don't get grouped together."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Timeout'
    self.testcases[0].crash_state = 'abcdef'
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'TimeoutX'
    self.testcases[1].crash_state = 'abcde'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)

  def test_different_project_name_with_similar_state(self):
    """Test that the crashes with different project names and similar crash
    state don't get grouped together."""
    self.testcases[0].security_flag = False
    self.testcases[0].crash_type = 'Heap-buffer-overflow'
    self.testcases[0].crash_state = 'abcdef'
    self.testcases[0].project_name = 'project1'
    self.testcases[1].security_flag = False
    self.testcases[1].crash_type = 'Heap-buffer-overflow'
    self.testcases[1].crash_state = 'abcde'
    self.testcases[1].project_name = 'project2'

    for t in self.testcases:
      t.put()

    grouper.group_testcases()

    for index, t in enumerate(self.testcases):
      self.testcases[index] = data_handler.get_testcase_by_id(t.key.id())
      self.assertEqual(self.testcases[index].group_id, 0)
      self.assertTrue(self.testcases[index].is_leader)


@test_utils.with_cloud_emulators('datastore')
class GroupExceedMaxTestcasesTest(unittest.TestCase):
  """Grouper test when a group exceeds maximum number of testcases."""

  def test_group_exceed_max_testcases(self):
    """Test that group auto-shrinks when it exceeds maximum number of
    testcases."""
    for i in range(1, 31):
      testcase = test_utils.create_generic_testcase()
      testcase.crash_type = 'Heap-buffer-overflow'
      testcase.crash_state = 'abcdefgh' + str(i)
      testcase.project_name = 'project'
      testcase.one_time_crasher_flag = False

      # Attach actual issues to some testcases.
      if i in [3, 4, 5]:
        testcase.bug_information = '123'

      # Make some testcases unreproducible.
      if i in [1, 2, 3]:
        testcase.one_time_crasher_flag = True

      testcase.put()

    unrelated_testcase = test_utils.create_generic_testcase()

    grouper.group_testcases()

    testcase_ids = list(data_handler.get_open_testcase_id_iterator())

    # [1, 2] get removed since they are unreproducible testcases.
    # [3] is not removed since it has bug attached (even though unreproducible).
    # [6, 7, 8] are removed to account for max group size. Even though they
    # are reproducible, they are the ones with least weight.
    expected_testcase_ids = [3, 4, 5] + list(range(
        9, 31)) + [unrelated_testcase.key.id()]
    self.assertEqual(expected_testcase_ids, testcase_ids)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Issue tracker interface."""


class LabelStore(object):
  """Label storage which tracks changes. Case insensitive, but preserves
  case."""

  def __init__(self, seq=()):
    self._backing = {}
    self._added = {}
    self._removed = {}

    for item in seq:
      self._backing[item.lower()] = item

  def __iter__(self):
    for value in self._backing.values():
      yield value

  def __contains__(self, item):
    return item.lower() in self._backing

  @property
  def added(self):
    return self._added.values()

  @property
  def removed(self):
    return self._removed.values()

  def add(self, label):
    """Add a new label."""
    if not label:
      return

    label = str(label)
    key = label.lower()
    if key in self._removed:
      del self._removed[key]
    else:
      self._added[key] = label

    self._backing[key] = label

  def remove(self, label):
    """Remove a label."""
    label = str(label)
    key = label.lower()
    if key not in self._backing:
      return

    if key in self._added:
      del self._added[key]
    else:
      self._removed[key] = label

    del self._backing[key]

  def reset_tracking(self):
    """Reset tracking."""
    self._added.clear()
    self._removed.clear()

  def clear(self):
    """Clear the store, and reset tracking."""
    self.reset_tracking()
    self._backing.clear()

  def get_by_prefix(self, prefix):
    """Get labels with the given prefix."""
    for item in self:
      if item.lower().startswith(prefix.lower()):
        yield item

  def has_with_prefix(self, prefix):
    """Return whether if there is an item with the given prefix."""
    return bool(next(self.get_by_prefix(prefix), None))

  def remove_by_prefix(self, prefix):
    """Remove labels with a given prefix."""
    for item in list(self):
      if item.lower().startswith(prefix.lower()):
        self.remove(item)

  def get_by_pattern(self, re_pattern):
    """Get labels with the given pattern."""
    for item in self:
      if re_pattern.match(item):
        yield item

  def has_with_pattern(self, re_pattern):
    """Return whether if there is an item with the given pattern."""
    return bool(next(self.get_by_pattern(re_pattern), None))


class Issue(object):
  """Represents an issue."""

  @property
  def issue_tracker(self):
    """The issue tracker for this issue."""
    raise NotImplementedError

  @property
  def id(self):
    """The issue identifier."""
    raise NotImplementedError

  @property
  def title(self):
    """The issue title."""
    raise NotImplementedError

  @title.setter
  def title(self, new_title):
    raise NotImplementedError

  @property
  def reporter(self):
    """The issue reporter."""
    raise NotImplementedError

  @reporter.setter
  def reporter(self, new_reporter):
    raise NotImplementedError

  @property
  def merged_into(self):
    """The issue that this is merged into."""
    raise NotImplementedError

  @property
  def closed_time(self):
    """When the issue was closed."""
    raise NotImplementedError

  @property
  def is_open(self):
    """Whether the issue is open."""
    raise NotImplementedError

  @property
  def status(self):
    """The issue status."""
    raise NotImplementedError

  @status.setter
  def status(self, new_status):
    raise NotImplementedError

  @property
  def body(self):
    """The issue body."""
    raise NotImplementedError

  @body.setter
  def body(self, new_body):
    raise NotImplementedError

  @property
  def assignee(self):
    """The issue assignee."""
    raise NotImplementedError

  @assignee.setter
  def assignee(self, new_assignee):
    raise NotImplementedError

  @property
  def ccs(self):
    """The issue CC list."""
    raise NotImplementedError

  @property
  def labels(self):
    """The issue labels list."""
    raise NotImplementedError

  @property
  def components(self):
    """The issue component list."""
    raise NotImplementedError

  @property
  def actions(self):
    """Get the issue actions."""
    raise NotImplementedError

  def save(self, new_comment=None, notify=True):
    """Save the issue."""
    raise NotImplementedError


class ChangeList(object):
  """Records a change in a list."""

  def __init__(self):
    self.added = []
    self.removed = []


class Action(object):
  """Represents an action on an issue (e.g. a comment)."""

  @property
  def author(self):
    """The author of the action."""
    raise NotImplementedError

  @property
  def comment(self):
    """Represents a comment."""
    raise NotImplementedError

  @property
  def title(self):
    """The new issue title."""
    raise NotImplementedError

  @property
  def status(self):
    """The new issue status."""
    raise NotImplementedError

  @property
  def assignee(self):
    """The new issue assignee."""
    raise NotImplementedError

  @property
  def ccs(self):
    """The issue CC change list."""
    raise NotImplementedError

  @property
  def labels(self):
    """The issue labels change list."""
    raise NotImplementedError

  @property
  def components(self):
    """The issue component change list."""
    raise NotImplementedError


class IssueTracker(object):
  """Issue tracker interface."""

  @property
  def project(self):
    """Get the project name of this issue tracker."""
    raise NotImplementedError

  @property
  def label_type(self):
    """Label type."""
    return 'label'  # default

  def label_text(self, label):
    """Text for a label (with label type)."""
    return label + ' ' + self.label_type

  def new_issue(self):
    """Create an unsaved new issue."""
    raise NotImplementedError

  def get_issue(self, issue_id):
    """Get the issue with the given ID."""
    raise NotImplementedError

  def get_original_issue(self, issue_id):
    """Retrieve the original issue object traversing the list of duplicates."""
    # Caller might pass |issue_id| as int, so change it to str so that
    # circular chain checks in loop actually work.
    original_issue_id = str(issue_id)
    seen_issue_ids = []
    while True:
      original_issue = self.get_issue(original_issue_id)
      if not original_issue:
        return None

      seen_issue_ids.append(original_issue_id)

      if not original_issue.merged_into:
        # If this is an original issue, no more work to do. Bail out.
        break

      original_issue_id = str(original_issue.merged_into)
      if original_issue_id in seen_issue_ids:
        # Don't traverse a circular chain, break if we realise that.
        break

    return original_issue

  def find_issues(self, keywords=None, only_open=None):
    """Find issues."""
    raise NotImplementedError

  def find_issues_url(self, keywords=None, only_open=None):
    """Find issues (web URL)."""
    raise NotImplementedError

  def issue_url(self, issue_id):
    """Return the issue URL with the given ID."""
    raise NotImplementedError



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Issue tracker manager functions."""

import datetime

from googleapiclient import discovery

from clusterfuzz._internal.base import retry

from . import credential_storage
from .comment import Comment
from .issue import ChangeList
from .issue import Issue

# Default value for issue tracker connection failures.
FAIL_RETRIES = 7
FAIL_WAIT = 1


def convert_entry_to_comment(entry):
  """Convert an issue entry object into a comment object."""
  comment = Comment()
  comment.author = entry['author']['name'] if 'author' in entry else None
  comment.comment = entry['content']
  comment.created = parse_datetime(entry['published'])
  comment.id = entry['id']

  if 'updates' in entry and entry['updates']:
    comment.cc = ChangeList(entry['updates'].get('cc', []))
    comment.components = ChangeList(entry['updates'].get('components', []))
    comment.labels = ChangeList(entry['updates'].get('labels', []))
    comment.owner = entry['updates'].get('owner', None)
    comment.status = entry['updates'].get('status', None)
    comment.summary = entry['updates'].get('summary', None)

  return comment


def convert_entry_to_issue(entry, itm, old_issue=None):
  """Convert an issue entry object into a issue object."""
  if old_issue:
    issue = old_issue
  else:
    issue = Issue()

  issue.blocked_on = [e['issueId'] for e in entry.get('blockedOn', [])]
  issue.blocking = [e['issueId'] for e in entry.get('blocking', [])]
  issue.cc = ChangeList([e['name'] for e in entry.get('cc', [])])
  issue.comments = None
  issue.components = ChangeList(entry.get('components', []))
  issue.created = parse_datetime(entry['published'])
  issue.id = entry['id']
  issue.itm = itm
  issue.labels = ChangeList(entry.get('labels', []))
  issue.new = False
  issue.open = entry['state'] == 'open'
  issue.reporter = entry['author']['name'] if 'author' in entry else None
  issue.stars = entry['stars']
  issue.summary = entry['summary']
  issue.updated = parse_datetime(entry['updated'])

  if entry.get('closed', []):
    issue.closed = parse_datetime(entry.get('closed', []))
  if entry.get('mergedInto'):
    issue.merged_into = entry['mergedInto'].get('issueId')
    issue.merged_into_project = entry['mergedInto'].get('projectId')
  if entry.get('owner', []):
    issue.owner = entry['owner']['name']
  if entry.get('status', []):
    issue.status = entry['status']

  # The issue will be flagged as dirty when most of the above fields are set,
  # so this must be set last.
  issue.dirty = False

  return issue


def parse_datetime(date_string):
  """Parse a date time string into a datetime object."""
  datetime_obj, _, microseconds_string = date_string.partition('.')
  datetime_obj = datetime.datetime.strptime(datetime_obj, '%Y-%m-%dT%H:%M:%S')
  if microseconds_string:
    microseconds = int(microseconds_string.rstrip('Z'), 10)
    return datetime_obj + datetime.timedelta(microseconds=microseconds)

  return datetime_obj


class IssueTrackerManager(object):
  """Issue tracker manager."""

  CAN_ALL = 'all'
  CAN_OPEN = 'open'
  CAN_MY_OPEN_BUGS = 'owned'
  CAN_REPORTED_BY_ME = 'reported'
  CAN_STARRED_BY_ME = 'starred'
  CAN_NEW = 'new'
  CAN_VERIFY = 'to-verify'

  CAN_VALUE_TO_ID_MAP = {
      CAN_ALL: 1,
      CAN_OPEN: 2,
      CAN_MY_OPEN_BUGS: 3,
      CAN_REPORTED_BY_ME: 4,
      CAN_STARRED_BY_ME: 5,
      CAN_NEW: 6
  }

  API_DISCOVERY_URL = ('https://monorail-prod.appspot.com/_ah/api/discovery/'
                       'v1/apis/{api}/{apiVersion}/rest')
  API_NAME = 'monorail'
  API_VERSION = 'v1'

  def __init__(self, project_name):
    """"Construct an issue tracker manager instance based on parameters."""
    self._client = None
    self.project_name = project_name

  @property
  def client(self):
    """HTTP Client."""
    if self._client is None:
      self._client = self._create_client()

    return self._client

  @retry.wrap(
      retries=FAIL_RETRIES,
      delay=FAIL_WAIT,
      function='libs.issue_management.issue_tracker_manager._execute_with_retry'
  )
  def _execute_with_retry(self, query):
    """Execute a query (with retries)."""
    return query.execute()

  @retry.wrap(
      retries=FAIL_RETRIES,
      delay=FAIL_WAIT,
      function='libs.issue_management.issue_tracker_manager._create_client')
  def _create_client(self):
    """Return a client object for querying the issue tracker.

    Includes retry logic to handle occasional 503 backend errors.
    """
    # Pull credentials old-style, using native client id and json.
    credentials = credential_storage.CredentialStorage().get()

    return discovery.build(
        self.API_NAME,
        self.API_VERSION,
        credentials=credentials,
        discoveryServiceUrl=self.API_DISCOVERY_URL,
        cache_discovery=False)

  def save(self, issue, send_email=None):
    """Save an issue and optionally send update notification over email."""
    if send_email is None:
      send_email = issue.send_email

    if issue.new:
      self._create(issue)
    else:
      self._update(issue, send_email)

  def _create(self, issue, send_email=True):
    """Create an issue and optionally send update notification over email."""
    cc = [{'name': user} for user in issue.cc]
    body = {
        'cc': cc,
        'components': issue.components,
        'description': issue.body,
        'labels': issue.labels,
        'projectId': self.project_name,
        'status': issue.status,
        'summary': issue.summary,
    }
    if issue.owner:
      body['owner'] = {'name': issue.owner}

    tmp = self._execute_with_retry(self.client.issues().insert(
        projectId=self.project_name, sendEmail=send_email, body=body))
    issue.id = int(tmp['id'])
    issue.dirty = False
    issue.new = False
    return issue

  def _update(self, issue, send_email=True):
    """Update an issue and optionally send update notification over email."""
    if not issue.dirty:
      return issue
    if not issue.owner:
      issue.owner = ''

    updates = {}
    if 'summary' in issue.changed:
      updates['summary'] = issue.summary
    if 'status' in issue.changed:
      updates['status'] = issue.status
    if 'owner' in issue.changed:
      updates['owner'] = issue.owner
    if issue.labels.is_changed():
      updates['labels'] = list(issue.labels.added)
    if issue.components.is_changed():
      updates['components'] = list(issue.components.added)
    if issue.cc.is_changed():
      updates['cc'] = list(issue.cc.added)

    body = {'id': issue.id, 'updates': updates}
    if 'comment' in issue.changed:
      body['content'] = issue.comment

    self._execute_with_retry(self.client.issues().comments().insert(
        projectId=self.project_name,
        issueId=issue.id,
        sendEmail=send_email,
        body=body))

    # Clear the issue comment once it's been saved (shouldn't be re-used).
    issue.comment = ''
    issue.dirty = False
    return issue

  def add_comment(self, issue_id, comment, send_email=True):
    """Add comment to an issue and potentially send an email update."""
    issue = self.get_issue(issue_id)
    issue.comment = comment
    self.save(issue, send_email)

  def get_comment_count(self, issue_id):
    """Get number of comments for an issue."""
    feed = self._execute_with_retry(self.client.issues().comments().list(
        projectId=self.project_name,
        issueId=issue_id,
        startIndex=1,
        maxResults=0))
    return feed.get('totalResults', '0')

  def get_comments(self, issue_id):
    """Get all comments for an issue."""
    comments = []
    comments_feed = self._execute_with_retry(
        self.client.issues().comments().list(
            projectId=self.project_name, issueId=issue_id))
    comments.extend(
        [convert_entry_to_comment(entry) for entry in comments_feed['items']])
    total_results = comments_feed['totalResults']
    if total_results:
      total_results = comments_feed['totalResults']
    else:
      return comments

    while len(comments) < total_results:
      comments_feed = self._execute_with_retry(
          self.client.issues().comments().list(
              projectId=self.project_name,
              issueId=issue_id,
              startIndex=len(comments)))
      comments.extend(
          [convert_entry_to_comment(entry) for entry in comments_feed['items']])

    return comments

  def get_first_comment(self, issue_id):
    """Get first comment for an issue."""
    feed = self._execute_with_retry(self.client.issues().comments().list(
        projectId=self.project_name,
        issueId=issue_id,
        startIndex=0,
        maxResults=1))
    if 'items' in feed:
      return convert_entry_to_comment(feed['items'][0])

    return None

  def get_last_comment(self, issue_id):
    """Get last comment for an issue."""
    total_results = self.get_comment_count(issue_id)
    feed = self._execute_with_retry(self.client.issues().comments().list(
        projectId=self.project_name,
        issueId=issue_id,
        startIndex=total_results - 1,
        maxResults=1))
    if 'items' in feed:
      return convert_entry_to_comment(feed['items'][0])

    return None

  def get_issue(self, issue_id):
    """Retrieve an issue object with a specific id."""
    entry = self._execute_with_retry(self.client.issues().get(
        projectId=self.project_name, issueId=issue_id))
    return convert_entry_to_issue(entry, self)

  def refresh(self, issue):
    """Refresh an issue object with latest updates."""
    if issue and not issue.new:
      entry = self._execute_with_retry(self.client.issues().get(
          projectId=self.project_name, issueId=issue.id))
      return convert_entry_to_issue(entry, self, old_issue=issue)

    return issue

  def get_all_issues(self):
    """Get all issues for the project."""
    feed = self._execute_with_retry(
        self.client.issues().list(projectId=self.project_name))
    return [convert_entry_to_issue(entry, self) for entry in feed['items']]

  def get_issue_count(self, query_string, can=CAN_ALL):
    """Return number of issues for a given query."""
    feed = self._execute_with_retry(self.client.issues().list(
        can=can,
        projectId=self.project_name,
        q=query_string,
        startIndex=0,
        maxResults=0))
    total_results = feed.get('totalResults', '')
    if total_results:
      return int(total_results)

    return 0

  def get_issues(self, query_string, can=CAN_ALL, max_results=1000):
    """Return all issues for a given query."""
    block_count = 0
    count = 0
    issues = []
    while True:
      result, total = self.get_issues_from_index(
          query_string, can, max_results=max_results, start_index=count)
      count += len(result)

      # Hack, since issue tracker is omitting results randomly.
      block_count += max_results
      issues += result
      if block_count > total:
        break

    return issues

  def get_issues_from_index(self,
                            query_string,
                            can=CAN_ALL,
                            max_results=1000,
                            start_index=0):
    """Retrieve a set of issues for a query from a given start index."""
    feed = self._execute_with_retry(self.client.issues().list(
        projectId=self.project_name,
        q=query_string,
        startIndex=start_index,
        maxResults=max_results,
        can=can))

    if 'items' in feed and feed['items']:
      issues = [convert_entry_to_issue(entry, self) for entry in feed['items']]
      return issues, feed['totalResults']

    return [], 0



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test base class for tasks which involve looking into component revisions."""

import ast
import os
import unittest

from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils

DATA_DIRECTORY = os.path.join(
    os.path.dirname(__file__), 'component_related_test_data')


@test_utils.with_cloud_emulators('datastore')
class ComponentRevisionPatchingTest(unittest.TestCase):
  """Base class for tests involving revisions of components;
  patches the function which retrieves those revisions."""

  def setUp(self):
    helpers.patch_environ(self)

    helpers.patch(self, [
        'clusterfuzz._internal.build_management.revisions.get_component_revisions_dict',
    ])

    self.mock.get_component_revisions_dict.side_effect = (
        self.mock_get_component_revisions_dict)

  @staticmethod
  def mock_get_component_revisions_dict(revision, _):
    if revision == 0:
      return {}

    component_revisions_file_path = os.path.join(
        DATA_DIRECTORY, 'component_revisions_%s.txt' % revision)
    with open(component_revisions_file_path) as file_handle:
      return ast.literal_eval(file_handle.read())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzz_strategy_selection cron job."""
# pylint: disable=protected-access

import json
import os
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import fuzz_strategy_selection

DATA_DIRECTORY = os.path.join(
    os.path.dirname(__file__), 'fuzz_strategy_selection_data')


@test_utils.with_cloud_emulators('datastore')
class TestFuzzStrategySelection(unittest.TestCase):
  """Tests whether the program properly stores
  calculated multi-armed bandit probabilities in datastore."""

  def setUp(self):
    """Set up method strategy distribution calculation tests."""
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'handlers.cron.fuzz_strategy_selection.'
        '_query_multi_armed_bandit_probabilities',
        'handlers.cron.fuzz_strategy_selection.'
        '_store_probabilities_in_bigquery'
    ])
    self.mock._query_multi_armed_bandit_probabilities.return_value = json.load(
        open(os.path.join(DATA_DIRECTORY, 'multi_armed_bandit_query.json')))

  def test_strategy_probabilities(self):
    """Ensure that the expected probabilities are being set for
    various methods."""

    # We pass an empty list as the query is patched out.
    fuzz_strategy_selection._query_and_upload_strategy_probabilities(
        fuzz_strategy_selection.LIBFUZZER_ENGINE)
    row1 = data_types.FuzzStrategyProbability.query(
        data_types.FuzzStrategyProbability.strategy_name ==
        'corpus_mutations_ml_rnn,fork,').get()
    self.assertEqual(row1.probability, 0.0061753988734031835)
    self.assertEqual(row1.engine, 'libFuzzer')
    row2 = data_types.FuzzStrategyProbability.query(
        data_types.FuzzStrategyProbability.strategy_name ==
        'corpus_mutations_ml_rnn,fork,corpus_subset,').get()
    self.assertEqual(row2.probability, 0.005077124089195762)
    self.assertEqual(row2.engine, 'libFuzzer')
    row3 = data_types.FuzzStrategyProbability.query(
        data_types.FuzzStrategyProbability.strategy_name ==
        'random_max_len,corpus_mutations_ml_rnn,recommended_dict,').get()
    self.assertEqual(row3.probability, 0.011855451527667878)
    self.assertEqual(row3.engine, 'libFuzzer')

  def test_delete_from_table(self):
    """Ensures that ndb datastore table is properly being
    cleared before being updated."""

    # We pass an empty list as the query is patched out.
    fuzz_strategy_selection._query_and_upload_strategy_probabilities(
        fuzz_strategy_selection.LIBFUZZER_ENGINE)
    count1 = data_types.FuzzStrategyProbability.query().count()
    fuzz_strategy_selection._query_and_upload_strategy_probabilities(
        fuzz_strategy_selection.LIBFUZZER_ENGINE)
    count2 = data_types.FuzzStrategyProbability.query().count()
    self.assertEqual(count1, count2)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for persistent_cache."""

import datetime

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class PersistentCacheTest(fake_filesystem_unittest.TestCase):
  """Tests for persistent_cache functions."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    environment.set_value('CACHE_DIR', '/tmp/test-cache')
    persistent_cache.initialize()

  def test_set_get_string(self):
    """Ensure it works with string value."""
    persistent_cache.set_value('key', 'value')
    self.assertEqual(persistent_cache.get_value('key'), 'value')

  def test_set_get_datetime(self):
    """Ensure it works with datetime value."""
    epoch = datetime.datetime.utcfromtimestamp(0)
    end_time = datetime.datetime.utcfromtimestamp(10)
    diff_time = end_time - epoch
    persistent_cache.set_value('key', diff_time.total_seconds())
    self.assertEqual(
        persistent_cache.get_value(
            'key', constructor=datetime.datetime.utcfromtimestamp), end_time)

  def test_delete(self):
    """Ensure it deletes key."""
    persistent_cache.set_value('key', 'value')
    persistent_cache.delete_value('key')
    self.assertEqual(persistent_cache.get_value('key'), None)

  def test_get_nonexistence(self):
    """Ensure it returns default_value when key doesn't exists."""
    self.assertEqual(persistent_cache.get_value('key', default_value=1), 1)

  def test_get_invalid(self):
    """Ensure it returns default_value when constructor fails."""
    time_now = datetime.datetime.utcnow()
    persistent_cache.set_value('key', 'random')
    self.assertEqual(persistent_cache.get_value('key'), 'random')
    self.assertEqual(
        persistent_cache.get_value(
            'key',
            default_value=time_now,
            constructor=datetime.datetime.utcfromtimestamp), time_now)

  def test_persist_across_reboots(self):
    """Ensure persist_across_reboots works."""
    persistent_cache.set_value('key', 'a')
    persistent_cache.set_value('key2', 'b', persist_across_reboots=True)
    persistent_cache.clear_values()
    self.assertEqual(persistent_cache.get_value('key'), None)
    self.assertEqual(persistent_cache.get_value('key2'), 'b')

  def test_clear_all(self):
    """Ensure clear all works."""
    persistent_cache.set_value('key', 'a')
    persistent_cache.set_value('key2', 'b', persist_across_reboots=True)
    persistent_cache.clear_values(clear_all=True)
    self.assertEqual(persistent_cache.get_value('key'), None)
    self.assertEqual(persistent_cache.get_value('key2'), None)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for app functions."""

import os
import unittest

from clusterfuzz._internal.platforms.android import settings
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'settings_data')


def _read_data_file(filename):
  return open(os.path.join(DATA_PATH, filename)).read()


class GetDeviceCodenameTest(unittest.TestCase):
  """Tests for get_device_codename."""

  def setUp(self):
    test_helpers.patch(
        self, ['clusterfuzz._internal.platforms.android.adb.run_command'])
    test_helpers.patch_environ(self)

    output = _read_data_file('get_device_codename_output.txt')
    self.mock.run_command.return_value = output

  def test_by_serial(self):
    """Ensure that we report the correct codename for serial number."""
    os.environ['ANDROID_SERIAL'] = '123456789012'
    self.assertEqual(settings.get_device_codename(), 'device1')

  def test_by_usb(self):
    """Ensure that we report the correct codename for a usb device."""
    os.environ['ANDROID_SERIAL'] = 'usb:2-4.2'
    self.assertEqual(settings.get_device_codename(), 'device2')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Monorail issue tracker."""

import urllib.parse

from libs.issue_management import issue_tracker
from libs.issue_management.monorail.issue import ChangeList
from libs.issue_management.monorail.issue import Issue as MonorailIssue
from libs.issue_management.monorail.issue_tracker_manager import (
    IssueTrackerManager)

ISSUE_TRACKER_URL = (
    'https://bugs.chromium.org/p/{project}/issues/detail?id={id}')
ISSUE_TRACKER_SEARCH_URL = (
    'https://bugs.chromium.org/p/{project}/issues/list?{params}')


class Issue(issue_tracker.Issue):
  """Represents an issue."""

  def __init__(self, monorail_issue):
    self._monorail_issue = monorail_issue

    # These mirror the underlying MonorailIssue data structures, to make it more
    # opaque to the client about how issue updates are done. For instance, when
    # a `label` is removed, what actually happens is `-label` is added. This
    # should not be visible to the client.
    self._ccs = issue_tracker.LabelStore(self._monorail_issue.cc)
    self._components = issue_tracker.LabelStore(self._monorail_issue.components)
    self._labels = issue_tracker.LabelStore(self._monorail_issue.labels)

  @property
  def issue_tracker(self):
    """The IssueTracker for this issue."""
    return IssueTracker(self._monorail_issue.itm)

  @property
  def id(self):
    """The issue identifier."""
    return self._monorail_issue.id

  @property
  def title(self):
    """The issue title."""
    return self._monorail_issue.summary

  @title.setter
  def title(self, new_title):
    self._monorail_issue.summary = new_title

  @property
  def reporter(self):
    """The issue reporter."""
    return self._monorail_issue.reporter

  @reporter.setter
  def reporter(self, new_reporter):
    self._monorail_issue.reporter = new_reporter

  @property
  def merged_into(self):
    """The issue that this is merged into."""
    if self._monorail_issue.merged_into_project != self.issue_tracker.project:
      # Don't consider duplicates in a different issue project.
      return None

    return self._monorail_issue.merged_into

  @property
  def is_open(self):
    """Whether the issue is open."""
    return self._monorail_issue.open

  @property
  def closed_time(self):
    return self._monorail_issue.closed

  @property
  def status(self):
    """The issue status."""
    return self._monorail_issue.status

  @status.setter
  def status(self, new_status):
    self._monorail_issue.status = new_status

  @property
  def body(self):
    """The issue body."""
    return self._monorail_issue.body

  @body.setter
  def body(self, new_body):
    self._monorail_issue.body = new_body

  @property
  def assignee(self):
    """The issue assignee."""
    return self._monorail_issue.owner

  @assignee.setter
  def assignee(self, new_assignee):
    self._monorail_issue.owner = new_assignee

  @property
  def ccs(self):
    """The issue CC list."""
    return self._ccs

  @property
  def labels(self):
    """The issue labels list."""
    return self._labels

  @property
  def components(self):
    """The issue component list."""
    return self._components

  @property
  def actions(self):
    """Get the issue actions."""
    return (Action(comment) for comment in self._monorail_issue.get_comments())

  def save(self, new_comment=None, notify=True):
    """Save the issue."""

    # Apply actual label changes to the underlying MonorailIssue.
    for added in self._components.added:
      self._monorail_issue.add_component(added)
    for removed in self._components.removed:
      self._monorail_issue.remove_component(removed)
    self._components.reset_tracking()

    for added in self._ccs.added:
      self._monorail_issue.add_cc(added)
    for removed in self._ccs.removed:
      self._monorail_issue.remove_cc(removed)
    self._ccs.reset_tracking()

    for added in self._labels.added:
      self._monorail_issue.add_label(added)
    for removed in self._labels.removed:
      self._monorail_issue.remove_label(removed)
    self._labels.reset_tracking()

    if new_comment:
      self._monorail_issue.comment = new_comment

    self._monorail_issue.save(send_email=notify)


class Action(issue_tracker.Action):
  """Monorail Action."""

  def __init__(self, monorail_comment):
    self._monorail_comment = monorail_comment

  @property
  def author(self):
    """The author of the action."""
    return self._monorail_comment.author

  @property
  def comment(self):
    """Represents a comment."""
    return self._monorail_comment.comment

  @property
  def title(self):
    """The new issue title."""
    return self._monorail_comment.summary

  @property
  def status(self):
    """The new issue status."""
    return self._monorail_comment.status

  @property
  def assignee(self):
    """The new issue assignee."""
    return self._monorail_comment.owner

  @property
  def ccs(self):
    """The issue CC change list."""
    return _to_change_list(self._monorail_comment.cc)

  @property
  def labels(self):
    """The issue labels change list."""
    return _to_change_list(self._monorail_comment.labels)

  @property
  def components(self):
    """The issue component change list."""
    return _to_change_list(self._monorail_comment.components)


class IssueTracker(issue_tracker.IssueTracker):
  """Issue tracker interface."""

  def __init__(self, itm):
    self._itm = itm

  @property
  def project(self):
    return self._itm.project_name

  def new_issue(self):
    monorail_issue = MonorailIssue()
    monorail_issue.itm = self._itm
    return Issue(monorail_issue)

  def get_issue(self, issue_id):
    monorail_issue = self._itm.get_issue(int(issue_id))
    if not monorail_issue:
      return None

    return Issue(monorail_issue)

  def find_issues(self, keywords=None, only_open=False):
    """Find issues."""
    if not keywords:
      return None

    search_text = _get_search_text(keywords)
    if only_open:
      can = IssueTrackerManager.CAN_OPEN
    else:
      can = IssueTrackerManager.CAN_ALL

    issues = self._itm.get_issues(search_text, can=can)
    return [Issue(issue) for issue in issues]

  def find_issues_url(self, keywords=None, only_open=False):
    """Find issues (web URL)."""
    if not keywords:
      return None

    search_text = _get_search_text(keywords)
    if only_open:
      can = IssueTrackerManager.CAN_OPEN
    else:
      can = IssueTrackerManager.CAN_ALL

    can_id = IssueTrackerManager.CAN_VALUE_TO_ID_MAP.get(can, '')
    return ISSUE_TRACKER_SEARCH_URL.format(
        project=self.project,
        params=urllib.parse.urlencode({
            'can_id': can_id,
            'q': search_text,
        }))

  def issue_url(self, issue_id):
    """Return the issue URL with the given ID."""
    return ISSUE_TRACKER_URL.format(project=self.project, id=issue_id)


def _to_change_list(monorail_list):
  """Convert a list of changed items to a issue_tracker.ChangeList."""
  change_list = issue_tracker.ChangeList()
  if not monorail_list:
    return change_list

  for item in monorail_list:
    if item.startswith('-'):
      change_list.removed.append(item[1:])
    else:
      change_list.added.append(item)

  return change_list


def _get_issue_tracker_manager_for_project(project_name):
  """Return monorail issue tracker manager for the given project."""
  # If there is no issue tracker set, bail out.
  if not project_name or project_name == 'disabled':
    return None

  return IssueTrackerManager(project_name=project_name)


def _get_search_text(keywords):
  """Get search text."""
  search_text = ' '.join(['"{}"'.format(keyword) for keyword in keywords])
  search_text = search_text.replace(':', ' ')
  search_text = search_text.replace('=', ' ')

  return search_text


def get_issue_tracker(project_name, config):  # pylint: disable=unused-argument
  """Get the issue tracker for the project name."""
  # TODO(ochang): Make this lazy.
  itm = _get_issue_tracker_manager_for_project(project_name)
  if itm is None:
    return None

  return IssueTracker(itm)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Load project data."""

from collections import namedtuple
import os

import six

from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.system import environment

FILE_SCHEME = 'file://'


class Project(object):
  """Project."""

  def __init__(self, project_id, clusters, instance_templates,
               host_worker_assignments):
    self.project_id = project_id
    self.clusters = clusters
    self.instance_templates = instance_templates
    self.host_worker_assignments = host_worker_assignments

  def get_instance_template(self, name):
    """Get instance template with the given name."""
    return next((template for template in self.instance_templates
                 if template['name'] == name), None)

  def get_cluster(self, name):
    """Get the cluster with the given name."""
    return next((cluster for cluster in self.clusters if cluster.name == name),
                None)


Cluster = namedtuple('Cluster', [
    'name', 'gce_zone', 'instance_count', 'instance_template', 'distribute',
    'worker', 'high_end'
])

HostWorkerAssignment = namedtuple('HostWorkerAssignment',
                                  ['host', 'worker', 'workers_per_host'])


def _process_instance_template(instance_template):
  """Process instance template, normalizing some of metadata key values."""
  # Load metadata items for a particular instance template.
  items = instance_template['properties']['metadata']['items']
  for item in items:
    # If the item value is a relative file path specified using the file://
    # scheme, then subtitute it with the actual file content. This is needed
    # since compute engine instance manager cannot read files from our repo.
    if (isinstance(item['value'], str) and
        item['value'].startswith(FILE_SCHEME)):
      file_path = item['value'][len(FILE_SCHEME):]
      with open(
          os.path.join(environment.get_gce_config_directory(), file_path)) as f:
        item['value'] = f.read()


def _config_to_project(name, config):
  """Read a project config."""
  clusters = []

  for cluster_name, zone in six.iteritems(config['clusters']):
    clusters.append(
        Cluster(
            name=cluster_name,
            gce_zone=zone['gce_zone'],
            instance_count=zone['instance_count'],
            instance_template=zone['instance_template'],
            distribute=zone.get('distribute', False),
            worker=zone.get('worker', False),
            high_end=zone.get('high_end', False)))

  for instance_template in config['instance_templates']:
    _process_instance_template(instance_template)

  host_worker_assignments = []
  for assignment in config.get('host_worker_assignments', []):
    host_worker_assignments.append(
        HostWorkerAssignment(
            host=assignment['host'],
            worker=assignment['worker'],
            workers_per_host=assignment['workers_per_host']))

  return Project(name, clusters, config['instance_templates'],
                 host_worker_assignments)


def _project_configs():
  return local_config.Config(local_config.GCE_CLUSTERS_PATH).get()


def get_projects():
  projects = []
  for name, project in six.iteritems(_project_configs()):
    projects.append(_config_to_project(name, project))

  return projects


def load_project(project_name):
  config = _project_configs().get(project_name)
  if not config:
    return None

  return _config_to_project(project_name, config)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""helper.py is a kitchen sink. It contains static methods that are used by
   multiple handlers."""

import logging
import sys
import traceback

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.datastore import data_handler
from libs import auth
from libs.issue_management import issue_tracker_utils

VIEW_OPERATION = 'View'
MODIFY_OPERATION = 'Modify'
TTL_IN_SECONDS = 15 * 60


class _DoNotCatchException(Exception):
  """Serve as a dummy exception to avoid catching any exception."""


class EarlyExitException(Exception):
  """Serve as an exception for exiting a handler's method early."""

  def __init__(self, message, status, trace_dump=None):
    super(EarlyExitException, self).__init__(message)
    self.status = status
    self.trace_dump = trace_dump
    if self.trace_dump is None:
      if sys.exc_info()[0] is not None:
        self.trace_dump = traceback.format_exc()
      else:
        self.trace_dump = ''.join(traceback.format_stack())

  def to_dict(self):
    """Build dict that is used for JSON serialisation."""
    return {
        'traceDump': self.trace_dump,
        'message': str(self),
        'email': get_user_email(),
        'status': self.status,
        'type': self.__class__.__name__
    }


class AccessDeniedException(EarlyExitException):
  """Serve as an exception for exiting a handler's method with 403."""

  def __init__(self, message=''):
    super(AccessDeniedException, self).__init__(message, 403, '')


class UnauthorizedException(EarlyExitException):
  """Serve as an exception for exiting a handler's method with 401."""

  def __init__(self, message=''):
    super(UnauthorizedException, self).__init__(message, 401, '')


def get_testcase(testcase_id):
  """Get a valid testcase or raise EarlyExitException."""
  testcase = None
  try:
    testcase = data_handler.get_testcase_by_id(testcase_id)
  except errors.InvalidTestcaseError:
    pass

  if not testcase:
    raise EarlyExitException("Testcase (id=%s) doesn't exist" % testcase_id,
                             404)
  return testcase


def get_issue_tracker_for_testcase(testcase):
  """Get an IssueTracker or raise EarlyExitException."""
  issue_tracker = issue_tracker_utils.get_issue_tracker_for_testcase(testcase)
  if not issue_tracker:
    raise EarlyExitException(
        "The testcase doesn't have a corresponding issue tracker", 404)
  return issue_tracker


def cast(value, fn, error_message):
  """Return `fn(value)` or raise an EarlyExitException with 400."""
  try:
    return fn(value)
  except (ValueError, TypeError):
    raise EarlyExitException(error_message, 400)


def should_render_json(accepts, content_type):
  """Check accepts and content_type to see if we should render JSON."""
  return 'application/json' in accepts or content_type == 'application/json'


def _is_not_empty(value):
  """Check if value is empty value or a tuple of empty values."""
  if isinstance(value, tuple):
    return any(bool(elem) for elem in value)

  return bool(value)


def get_or_exit(fn,
                not_found_message,
                error_message,
                not_found_exception=_DoNotCatchException,
                non_empty_fn=_is_not_empty):
  """Get an entity using `fn`. If the returning entity is nothing (e.g. None or
    a tuple on Nones), it raises 404.

    Args:
      fn: the function to get an entity. It's a function because fn(..) might
          raise an exception.
      not_found_message: the 404 HTTP error is raised with not_found_message for
          an empty entity.
      error_message: the 500 HTTP error is raised with error_message for any
          other exception from fn(..).
      not_found_exception: the type of exception that will be considered as
          'not found' as opposed to other errors."""
  result = None
  try:
    result = fn()
  except not_found_exception:
    pass
  except Exception:
    raise EarlyExitException(
        '%s (%s: %s)' % (error_message, sys.exc_info()[0], str(
            sys.exc_info()[1])), 500)

  if non_empty_fn(result):
    return result
  raise EarlyExitException(not_found_message, 404)


def get_user_email():
  """Returns currently logged-in user's email."""
  try:
    return auth.get_current_user().email
  except Exception:
    return ''


def get_integer_key(request):
  """Convenience function for getting an integer datastore key ID."""
  key = request.get('key')
  try:
    return int(key)
  except (ValueError, KeyError):
    raise EarlyExitException('Invalid key format.', 400)


def log(message, operation_type):
  """Logs operation being carried by current logged-in user."""
  logging.info('ClusterFuzz: %s (%s): %s.', operation_type, get_user_email(),
               message)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get values from the global configuration."""

import base64

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment

BASE64_MARKER = 'base64;'


def get():
  """Return configuration data."""
  # The reproduce tool does not have access to datastore. Rather than try to
  # catch all uses and handle them individually, we catch any accesses here.
  if environment.get_value('REPRODUCE_TOOL'):
    return None

  return data_types.Config.query().get()


def get_value(key):
  """Return a configuration key value."""
  config = get()
  if not config:
    return None

  value = config.__getattribute__(key)

  # Decode if the value is base64 encoded.
  if value.startswith(BASE64_MARKER):
    return base64.b64decode(value[len(BASE64_MARKER):])

  return value


def get_value_for_job(data, target_job_type):
  """Parses a value for a particular job type. If job type is not found,
  return the default value."""
  # All data is in a single line, just return that.
  if ';' not in data:
    return data

  result = ''
  for line in data.splitlines():
    job_type, value = (line.strip()).split(';')
    if job_type == target_job_type or (job_type == 'default' and not result):
      result = value

  return result


def set_value(key, value):
  """Sets a configuration key value and commits change."""
  config = get()
  if not config:
    return

  try:
    config.__setattr__(key, value)
  except UnicodeDecodeError:
    value = '%s%s' % (BASE64_MARKER, base64.b64encode(value))
    config.__setattr__(key, value)

  config.put()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""NDB initialization."""

import contextlib
import functools
import os
import threading

from google.cloud import ndb
from google.cloud.ndb import context as context_module

from clusterfuzz._internal.base import utils

_ndb_client = None
_ndb_client_lock = threading.Lock()
_initial_pid = None


def _client():
  """Get or initialize the NDB client."""
  global _ndb_client
  global _initial_pid

  if not _ndb_client:
    with _ndb_client_lock:
      if not _ndb_client:
        _ndb_client = ndb.Client(project=utils.get_application_id())
        _initial_pid = os.getpid()

  return _ndb_client


@contextlib.contextmanager
def context():
  """Get the NDB context."""
  if _initial_pid != os.getpid():
    # Forked, clear the existing context to avoid issues.
    # TODO(ochang): Remove this hack once on Python 3, where we can set
    # multiprocessing.set_start_method to not fork.
    context_module._state.context = None  # pylint: disable=protected-access

  with _client().context() as ndb_context:
    # Disable NDB caching, as NDB on GCE VMs do not use memcache and therefore
    # can't invalidate the memcache cache.
    ndb_context.set_memcache_policy(False)

    # Disable the in-context cache, as it can use up a lot of memory for
    # longer running tasks such as cron jobs.
    ndb_context.set_cache_policy(False)

    yield ndb_context


def thread_wrapper(func):
  """Wrapper for thread targets to initialize an NDB context, since contexts are
  thread local."""

  @functools.wraps(func)
  def _wrapper(*args, **kwargs):
    with context():
      return func(*args, **kwargs)

  return _wrapper



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the home page."""

import datetime
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import home
from libs import access
from libs import helpers


@test_utils.with_cloud_emulators('datastore')
class HomeTests(unittest.TestCase):
  """home tests."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.get_access',
        'clusterfuzz._internal.base.external_users.allowed_jobs_for_user',
        'libs.helpers.get_user_email',
    ])

    data_types.Job(
        name='libfuzzer_asan_lib',
        environment_string=('PROJECT_NAME = lib\n'
                            'CORPUS_PRUNE = True')).put()
    data_types.Job(
        name='afl_asan_lib', environment_string=('PROJECT_NAME = lib\n')).put()
    data_types.Job(
        name='libfuzzer_msan_lib',
        environment_string='PROJECT_NAME = lib').put()
    data_types.Job(
        name='afl_asan_lib2',
        environment_string=('PROJECT_NAME = lib2\n')).put()

    data_types.Job(
        name='libfuzzer_asan_lib2',
        environment_string=('PROJECT_NAME = lib2\n'
                            'CORPUS_PRUNE = True')).put()
    data_types.Job(
        name='libfuzzer_ubsan_lib2',
        environment_string='PROJECT_NAME = lib2').put()

    data_types.FuzzTarget(engine='afl', binary='fuzzer', project='lib2').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='afl_lib2_fuzzer',
        job='afl_asan_lib2',
        last_run=datetime.datetime.utcnow()).put()
    data_types.FuzzTarget(
        engine='libFuzzer', binary='fuzzer', project='lib2').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_lib2_fuzzer',
        job='libfuzzer_asan_lib2',
        last_run=datetime.datetime.utcnow()).put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_lib2_fuzzer',
        job='libfuzzer_ubsan_lib2',
        last_run=datetime.datetime.utcnow()).put()

    self.maxDiff = None  # pylint: disable=invalid-name

  def test_no_permissions(self):
    """Test user with no permissions."""
    self.mock.get_user_email.return_value = 'nope@nope.com'
    self.mock.allowed_jobs_for_user.return_value = []
    self.mock.get_access.return_value = access.UserAccess.Denied

    with self.assertRaises(helpers.EarlyExitException):
      home.get_results()

  def test_results_external(self):
    """Test results for external user."""
    self.mock.get_user_email.return_value = 'user@user.com'
    self.mock.allowed_jobs_for_user.return_value = [
        'libfuzzer_asan_lib', 'libfuzzer_msan_lib'
    ]
    self.mock.get_access.return_value = access.UserAccess.Denied

    results = home.get_results()
    self.assertEqual(
        results, {
            'info': {
                'projects': [{
                    'jobs': [{
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'libfuzzer_asan_lib',
                        'single_target': None,
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: memory (MSAN)',
                        'name': u'libfuzzer_msan_lib',
                        'single_target': None,
                        'engine_name': 'libFuzzer',
                        'engine_display_name': 'libFuzzer',
                        'has_stats': True,
                    }],
                    'name':
                        u'lib',
                    'coverage_job':
                        'libfuzzer_asan_lib',
                }],
                'is_internal_user':
                    False,
            }
        })

    self.mock.allowed_jobs_for_user.return_value = [
        'afl_asan_lib', 'libfuzzer_asan_lib', 'libfuzzer_msan_lib',
        'afl_asan_lib2', 'libfuzzer_asan_lib2', 'libfuzzer_ubsan_lib2'
    ]

    results = home.get_results()
    self.assertEqual(
        results, {
            'info': {
                'projects': [{
                    'jobs': [{
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'afl_asan_lib',
                        'single_target': None,
                        'engine_display_name': 'AFL',
                        'engine_name': 'afl',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'libfuzzer_asan_lib',
                        'single_target': None,
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: memory (MSAN)',
                        'name': u'libfuzzer_msan_lib',
                        'single_target': None,
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }],
                    'name':
                        u'lib',
                    'coverage_job':
                        'libfuzzer_asan_lib',
                }, {
                    'jobs': [{
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'afl_asan_lib2',
                        'single_target': 'afl_lib2_fuzzer',
                        'engine_display_name': 'AFL',
                        'engine_name': 'afl',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'libfuzzer_asan_lib2',
                        'single_target': 'libFuzzer_lib2_fuzzer',
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: undefined (UBSAN)',
                        'name': u'libfuzzer_ubsan_lib2',
                        'single_target': 'libFuzzer_lib2_fuzzer',
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }],
                    'name':
                        u'lib2',
                    'coverage_job':
                        'libfuzzer_asan_lib2',
                }],
                'is_internal_user':
                    False,
            }
        })

  def test_results_internal(self):
    """Test results for internal user."""
    self.mock.get_user_email.return_value = 'user@user.com'
    self.mock.allowed_jobs_for_user.return_value = []
    self.mock.get_access.return_value = access.UserAccess.Allowed

    results = home.get_results()
    self.assertEqual(
        results, {
            'info': {
                'projects': [{
                    'jobs': [{
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'afl_asan_lib',
                        'single_target': None,
                        'engine_display_name': 'AFL',
                        'engine_name': 'afl',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'libfuzzer_asan_lib',
                        'single_target': None,
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: memory (MSAN)',
                        'name': u'libfuzzer_msan_lib',
                        'single_target': None,
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }],
                    'name':
                        u'lib',
                    'coverage_job':
                        'libfuzzer_asan_lib',
                }, {
                    'jobs': [{
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'afl_asan_lib2',
                        'single_target': 'afl_lib2_fuzzer',
                        'engine_display_name': 'AFL',
                        'engine_name': 'afl',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: address (ASAN)',
                        'name': u'libfuzzer_asan_lib2',
                        'single_target': 'libFuzzer_lib2_fuzzer',
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }, {
                        'sanitizer_string': 'Sanitizer: undefined (UBSAN)',
                        'name': u'libfuzzer_ubsan_lib2',
                        'single_target': 'libFuzzer_lib2_fuzzer',
                        'engine_display_name': 'libFuzzer',
                        'engine_name': 'libFuzzer',
                        'has_stats': True,
                    }],
                    'name':
                        u'lib2',
                    'coverage_job':
                        'libfuzzer_asan_lib2',
                }],
                'is_internal_user':
                    True,
            }
        })



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler that schedules ML train jobs."""

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from clusterfuzz._internal.metrics import logs
from handlers import base_handler
from libs import handler

MODEL_NAME_TO_TASK = {
    'rnn_generator': 'train_rnn_generator',
}


class Handler(base_handler.Handler):
  """Schedule ML train tasks."""

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    for job in data_types.Job.query():

      models = job.get_environment().get('ML_MODELS_TO_USE')
      if not models:
        continue

      task_list = []
      for model_name in models.split(','):
        try:
          task_list.append(MODEL_NAME_TO_TASK[model_name.strip()])
        except KeyError:
          logs.log_error(f'Invalid ML model {model_name} for job {job.name}.')

      if not task_list:
        continue

      target_jobs = list(fuzz_target_utils.get_fuzz_target_jobs(job=job.name))
      fuzz_targets = fuzz_target_utils.get_fuzz_targets_for_target_jobs(
          target_jobs)

      for task_name in task_list:
        for target in fuzz_targets:
          tasks.add_task(
              task_name,
              target.fully_qualified_name(),
              job.name,
              queue=tasks.ML_JOBS_TASKQUEUE)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for performance_analyzer."""

import json
import os
import unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import libfuzzer
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    stats as performance_stats
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from handlers.performance_report import performance_analyzer

# Use default values for stats values usually provided by CF.
DEFAULT_STATS_PROVIDED_BY_CF = {
    'actual_duration': 2350,
    'expected_duration': 2350,
    'timestamp': 1499904000.017923
}


def _get_stats_from_log(log_path,
                        strategies=None,
                        arguments=None,
                        stats_overrides=None):
  """Calculate stats for the given log the same way as the engine does."""
  if strategies is None:
    strategies = []
  if arguments is None:
    arguments = []

  log_lines = utils.decode_to_unicode(
      utils.read_data_from_file(log_path, eval_data=False)).splitlines()
  stats = libfuzzer.parse_log_stats(log_lines)
  stats.update(
      performance_stats.parse_performance_features(log_lines, strategies,
                                                   arguments))
  if stats_overrides:
    stats.update(stats_overrides)

  return stats


class PerformanceAnalyzerTestBase(unittest.TestCase):
  """Performance analysis tests base class."""

  def setUp(self):
    """Prepare test data and necessary env variables."""
    test_helpers.patch_environ(self)
    self.data_directory = os.path.join(
        os.path.dirname(__file__), 'performance_analyzer_data')
    self.libfuzzer_data_directory = os.path.join(self.data_directory,
                                                 'libfuzzer')
    environment.set_value('FAIL_RETRIES', 1)

    self.analyzer = performance_analyzer.LibFuzzerPerformanceAnalyzer()


class PerformanceAnalyzerBasicAnalyzerTest(PerformanceAnalyzerTestBase):
  """Performance analysis tests for BasicAnalyzer functions."""

  def assert_basic_analyzer(self,
                            basic_analyzer,
                            log_filename,
                            stats_overrides=None):
    """Assert for testing a single basic analyzer."""
    extra_stats = DEFAULT_STATS_PROVIDED_BY_CF.copy()
    if stats_overrides:
      extra_stats.update(stats_overrides)

    log_file_path = os.path.join(self.libfuzzer_data_directory, 'issue_logs',
                                 log_filename)
    stats = _get_stats_from_log(log_file_path, stats_overrides=extra_stats)
    self.assertGreater(basic_analyzer(stats), 0.0)

  def test_basic_analyzer_for_bad_instrumentation(self):
    """Test analyzer_bad_instrumentation BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_bad_instrumentation,
                               'bad_instrumentation_issue.txt')

  def test_basic_analyzer_for_coverage(self):
    """Test analyzer_coverage BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_coverage,
                               'coverage_issue.txt')

  def test_basic_analyzer_for_crash(self):
    """Test analyzer_crash BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_crash, 'crash_issue.txt')

  def test_basic_analyzer_for_leak(self):
    """Test analyzer_leak BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_leak, 'leak_issue.txt')

  def test_basic_analyzer_for_logging(self):
    """Test analyzer_logging BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_logging,
                               'logging_issue.txt')

  def test_basic_analyzer_for_oom(self):
    """Test analyzer_oom BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_oom, 'oom_issue.txt')

  def test_basic_analyzer_for_slow_unit(self):
    """Test analyzer_slow_unit BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_slow_unit,
                               'slow_unit_issue.txt')

  def test_basic_analyzer_for_speed(self):
    """Test analyzer_speed BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_speed, 'speed_issue.txt')

  def test_basic_analyzer_for_startup_crash(self):
    """Test analyzer_startup_crash BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_startup_crash,
                               'startup_crash_issue.txt')

  def test_basic_analyzer_for_timeout(self):
    """Test analyzer_timeout BasicAnalyzer."""
    self.assert_basic_analyzer(self.analyzer.analyzer_timeout,
                               'timeout_issue.txt')


class PerformanceAnalyzerTest(PerformanceAnalyzerTestBase):
  """Performance analysis tests."""

  def get_issues(self, log_filename, stats_overrides=None):
    """Returns the issue for a particular log file."""
    extra_stats = DEFAULT_STATS_PROVIDED_BY_CF.copy()
    if stats_overrides:
      extra_stats.update(stats_overrides)

    log_file_path = os.path.join(self.libfuzzer_data_directory, 'issue_logs',
                                 log_filename)
    stats = _get_stats_from_log(log_file_path, stats_overrides=extra_stats)

    analyzer = performance_analyzer.LibFuzzerPerformanceAnalyzer()
    performance_scores, affected_runs_percents, examples = (
        analyzer.analyze_stats([stats]))

    return analyzer.get_issues(performance_scores, affected_runs_percents,
                               examples)

  def assert_log_has_issue_matching(self,
                                    log_filename,
                                    expected_issue,
                                    stats_overrides=None):
    """Assert for testing log has a particular issue."""
    detected_issues = self.get_issues(log_filename, stats_overrides)
    actual_issue = next(
        (i for i in detected_issues if i['type'] == expected_issue['type']),
        None)

    self.assertIsNotNone(
        actual_issue,
        '"%s" issue is not found in the result' % expected_issue['type'])
    self.assertEqual(actual_issue['percent'], expected_issue['percent'])
    self.assertEqual(actual_issue['score'], expected_issue['score'])

  def assert_log_has_no_issue_matching(self, log_filename, issue_type):
    """Assert for testing log has no issue matching a particular type."""
    detected_issues = self.get_issues(log_filename)
    detected_issue_types = [i['type'] for i in detected_issues]
    self.assertNotIn(issue_type, detected_issue_types)

  def assert_log_has_no_issues(self, log_filename, stats_overrides=None):
    """Assert for testing log has no issues."""
    expected_issue = {
        'type': 'none',
        'percent': 100.0,
        'score': 0.0,
        'examples': []
    }
    detected_issues = self.get_issues(log_filename, stats_overrides)
    self.assertEqual(detected_issues, [expected_issue])

  def test_bad_instrumentation(self):
    """Test bad instrumentation issue."""
    expected_issue = {
        'type': 'bad_instrumentation',
        'percent': 100.0,
        'score': 256.0
    }
    self.assert_log_has_issue_matching('bad_instrumentation_issue.txt',
                                       expected_issue)

  def test_coverage(self):
    """Test coverage issue."""
    expected_issue = {'type': 'coverage', 'percent': 100.0, 'score': 1.0}
    self.assert_log_has_issue_matching('coverage_issue.txt', expected_issue)

  def test_crash(self):
    """Test crash issue."""
    expected_issue = {'type': 'crash', 'percent': 100.0, 'score': 4.0}
    self.assert_log_has_issue_matching('crash_issue.txt', expected_issue)

  def test_leak(self):
    """Test leak issue."""
    expected_issue = {'type': 'leak', 'percent': 100.0, 'score': 128.0}
    self.assert_log_has_issue_matching('leak_issue.txt', expected_issue)

  def test_logging(self):
    """Test logging issue."""
    expected_issue = {'type': 'logging', 'percent': 100.0, 'score': 0.86}
    self.assert_log_has_issue_matching('logging_issue.txt', expected_issue)

  def test_oom(self):
    """Test oom issue."""
    expected_issue = {'type': 'oom', 'percent': 100.0, 'score': 8.0}
    self.assert_log_has_issue_matching('oom_issue.txt', expected_issue)

  def test_slow_unit(self):
    """Test slow unit issue."""
    expected_issue = {'type': 'slow_unit', 'percent': 100.0, 'score': 2.4}
    self.assert_log_has_issue_matching('slow_unit_issue.txt', expected_issue)

  def test_speed(self):
    """Test speed issue."""
    expected_issue = {'type': 'speed', 'percent': 100.0, 'score': 0.99}
    self.assert_log_has_issue_matching('speed_issue.txt', expected_issue)

  def test_startup_crash(self):
    """Test startup crash issue."""
    stats_overrides = {'average_exec_per_sec': 0, 'new_units_added': 0}
    expected_issue = {'type': 'startup_crash', 'percent': 100.0, 'score': 256.0}
    self.assert_log_has_issue_matching('startup_crash_issue.txt',
                                       expected_issue, stats_overrides)

  def test_no_startup_crash(self):
    """Test startup crash is not detected when libFuzzer exists early."""
    self.assert_log_has_no_issue_matching('startup_crash_no_issue.txt',
                                          'startup_crash')

  def test_timeout(self):
    """Test 'timeout' issue."""
    expected_issue = {'type': 'timeout', 'percent': 100.0, 'score': 64.0}
    self.assert_log_has_issue_matching('timeout_issue.txt', expected_issue)

  def test_no_logging_with_recommended_dictionaries(self):
    """Test no logging issue for a log with recommended dictionaries."""
    self.assert_log_has_no_issue_matching(
        'logging_recommended_dictionary_no_issue.txt', 'logging')

  def test_no_logging_with_crash(self):
    """Test no logging issue for a log with crash."""
    self.assert_log_has_no_issue_matching('logging_crash_no_issue.txt',
                                          'logging')

  def test_no_logging_with_slow_units(self):
    """Test no logging issue for a log with slow units."""
    self.assert_log_has_no_issue_matching('logging_slow_units_no_issue.txt',
                                          'logging')

  def test_no_logging_with_sanitizer_frames(self):
    """Test no logging issue for a log with sanitizer warnings."""
    self.assert_log_has_no_issue_matching(
        'logging_sanitizer_warnings_no_issue.txt', 'logging')

  def test_no_logging_with_oom(self):
    """Test no logging issue for a log with ooms in between frames."""
    self.assert_log_has_no_issue_matching('logging_oom_no_issue.txt', 'logging')

  def test_no_logging_with_few_runs(self):
    """Test no logging issue for a log with few runs."""
    self.assert_log_has_no_issue_matching('logging_few_runs_no_issue.txt',
                                          'logging')

  def test_corpus_subset_run_speed_coverage(self):
    """Test corpus subset run is ignored for a log with speed and coverage
    issue."""
    self.assert_log_has_no_issues('corpus_subset_no_coverage_speed_issue.txt')

  def test_corpus_subset_run_crash(self):
    """Test corpus subset run is not ignored for a log with crash."""
    expected_issue = {'type': 'crash', 'percent': 100.0, 'score': 4.0}
    self.assert_log_has_issue_matching('corpus_subset_crash_issue.txt',
                                       expected_issue)

  def test_perfect_fuzzer(self):
    """Test a perfect fuzzer, i.e. with no issues."""
    self.assert_log_has_no_issues('no_issue.txt')

  def test_report_generation(self):
    """Test report generation for a directory."""
    analyzer = performance_analyzer.LibFuzzerPerformanceAnalyzer()
    report_logs_directory = os.path.join(self.libfuzzer_data_directory,
                                         'report_logs')
    stats_rows = []

    # Use default values for stats values usually provided by CF.
    stats_overrides = DEFAULT_STATS_PROVIDED_BY_CF.copy()

    for filename in sorted(os.listdir(report_logs_directory)):
      # Use different timestamp values for each log.
      stats_overrides['timestamp'] += 1

      stats_rows.append(
          _get_stats_from_log(
              os.path.join(report_logs_directory, filename),
              stats_overrides=stats_overrides))

    performance_scores, affected_runs_percents, examples = (
        analyzer.analyze_stats(stats_rows))

    performance_issues = analyzer.get_issues(performance_scores,
                                             affected_runs_percents, examples)
    performance_report = performance_analyzer.generate_report(
        performance_issues, 'fuzzer1', 'job1')

    expected_report = utils.read_data_from_file(
        os.path.join(self.libfuzzer_data_directory, 'expected_report.json'),
        eval_data=False)

    self.maxDiff = None  # pylint: disable=invalid-name
    self.assertEqual(
        json.loads(performance_report), json.loads(expected_report))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pub/Sub helpers."""

import base64
import collections
import json
import threading

import googleapiclient
import httplib2

from clusterfuzz._internal.base import retry
from clusterfuzz._internal.google_cloud_utils import credentials
from clusterfuzz._internal.system import environment

MAX_ACK_DEADLINE = 10 * 60  # 10 minutes (the maximum).
_DEFAULT_MAX_MESSAGES = 1000  # Arbitrary reasonably large limit.
_PUBSUB_FAIL_RETRIES = 5
_PUBSUB_FAIL_WAIT = 2

Topic = collections.namedtuple('Topic', ['name', 'labels'])
Subscription = collections.namedtuple(
    'Subsciption', ['name', 'topic', 'ack_deadline', 'labels'])


class Message(object):
  """Pubsub message."""

  def __init__(self, data=None, attributes=None):
    self.data = data
    self.attributes = attributes


class ReceivedMessage(Message):
  """Received pubsub message."""

  def __init__(self, client, subscription, data, attributes, message_id,
               publish_time, ack_id):
    super().__init__(data, attributes)
    self._client = client
    self.ack_id = ack_id
    self.message_id = message_id
    self.subscription = subscription
    self.publish_time = publish_time

  def ack(self):
    """Acknowledge the message."""
    self._client.ack(self.subscription, [self.ack_id])

  def modify_ack_deadline(self, seconds):
    """Modify the acknowledgement deadline."""
    self._client.modify_ack_deadline(self.subscription, [self.ack_id], seconds)


class PubSubClient(object):
  """Helper class that provides wrappers around some Pub/Sub functionality."""

  def __init__(self, emulator_host=None):
    self._local = threading.local()
    self._emulator_host = emulator_host

  @retry.wrap(
      retries=_PUBSUB_FAIL_RETRIES,
      delay=_PUBSUB_FAIL_WAIT,
      function='google_cloud_utils.pubsub._api_client')
  def _api_client(self):
    """Get the client for the current thread."""
    if hasattr(self._local, 'api_client'):
      return self._local.api_client

    emulator_host = environment.get_value('PUBSUB_EMULATOR_HOST')
    creds = credentials.get_default()[0]
    if emulator_host:
      # Replace real discovery document's root url with the emulator.
      _, discovery_doc = httplib2.Http().request(
          googleapiclient.discovery.DISCOVERY_URI.format(
              api='pubsub', apiVersion='v1'))
      discovery_doc = json.loads(discovery_doc)
      discovery_doc['rootUrl'] = 'http://{}/'.format(emulator_host)

      self._local.api_client = googleapiclient.discovery.build_from_document(
          discovery_doc, credentials=creds)
    else:
      self._local.api_client = googleapiclient.discovery.build(
          'pubsub', 'v1', cache_discovery=False, credentials=creds)

    return self._local.api_client

  @retry.wrap(
      retries=_PUBSUB_FAIL_RETRIES,
      delay=_PUBSUB_FAIL_WAIT,
      function='google_cloud_utils.pubsub._execute_with_retry')
  def _execute_with_retry(self, request):
    """Execute a request (with retries)."""
    try:
      return request.execute()
    except googleapiclient.errors.HttpError as e:
      if e.resp.status == 404:
        return None

      raise

  def ack(self, subscription, ack_ids):
    """Acknowledge messages."""
    ack_body = {'ackIds': ack_ids}
    request = self._api_client().projects().subscriptions().acknowledge(
        subscription=subscription, body=ack_body)
    self._execute_with_retry(request)

  def modify_ack_deadline(self, subscription, ack_ids, seconds):
    """Modify acknowledgement deadline of messages."""
    body = {
        'ackIds': ack_ids,
        'ackDeadlineSeconds': int(seconds),  # since time.time() is float.
    }

    request = self._api_client().projects().subscriptions().modifyAckDeadline(
        subscription=subscription, body=body)
    self._execute_with_retry(request)

  def publish(self, topic, messages):
    """Publish a message to a topic."""
    request_body = {
        'messages': [_message_to_dict(message) for message in messages]
    }
    request = self._api_client().projects().topics().publish(
        topic=topic, body=request_body)

    response = self._execute_with_retry(request)
    if response is None:
      raise RuntimeError('Invalid topic: ' + topic)

    return sorted(response.get('messageIds'))

  def pull_from_subscription(self,
                             subscription,
                             max_messages=_DEFAULT_MAX_MESSAGES,
                             acknowledge=False):
    """Pull messages from a subscription to a topic."""
    request_body = {
        'returnImmediately': True,
        'maxMessages': max_messages,
    }
    request = self._api_client().projects().subscriptions().pull(
        subscription=subscription, body=request_body)
    response = self._execute_with_retry(request)
    if response is None:
      raise RuntimeError('Invalid subscription: ' + subscription)

    received_messages = response.get('receivedMessages')
    if not received_messages:
      # |receivedMessages| attribute may not exist if there are no messages in
      # the subscription. In that case, return an empty list.
      return []

    if acknowledge:
      ack_ids = [message['ackId'] for message in received_messages]
      self.ack(subscription, ack_ids)

    return [
        _raw_message_to_received_message(self, subscription, message)
        for message in received_messages
    ]

  def create_topic(self, name, labels=None):
    """Create a new topic."""
    body = {
        'labels': {},
    }

    if labels:
      body['labels'] = labels

    request = self._api_client().projects().topics().create(
        name=name, body=body)
    self._execute_with_retry(request)

  def delete_topic(self, name):
    """Deletes a topic."""
    request = self._api_client().projects().topics().delete(topic=name)
    self._execute_with_retry(request)

  def get_topic(self, name):
    """Get a topic, or None if it does not exist."""
    request = self._api_client().projects().topics().get(topic=name)
    response = self._execute_with_retry(request)
    if not response:
      return None

    return Topic(response['name'], response.get('labels'))

  def create_subscription(self,
                          name,
                          topic,
                          ack_deadline=MAX_ACK_DEADLINE,
                          labels=None):
    """Create a new subscription."""
    body = {'topic': topic, 'ackDeadlineSeconds': ack_deadline, 'labels': {}}

    if labels:
      body['labels'] = labels

    request = self._api_client().projects().subscriptions().create(
        name=name, body=body)
    self._execute_with_retry(request)

  def delete_subscription(self, name):
    """Delete a subscription."""
    request = self._api_client().projects().subscriptions().delete(
        subscription=name)
    self._execute_with_retry(request)

  def get_subscription(self, name):
    """Get a subscription, or None if it does not exist."""
    request = self._api_client().projects().subscriptions().get(
        subscription=name)
    response = self._execute_with_retry(request)
    if not response:
      return None

    return Subscription(response['name'], response['topic'],
                        response['ackDeadlineSeconds'], response.get('labels'))

  def list_topics(self, project, page_size=1000):
    """List topics."""
    request = self._api_client().projects().topics().list(
        project=project, pageSize=page_size)
    response = self._execute_with_retry(request)

    while True:
      for topic in response['topics']:
        yield topic['name']

      next_page_token = response.get('nextPageToken')
      if not next_page_token:
        break

      request = self._api_client().projects().topics().list(
          project=project, pageToken=next_page_token, pageSize=page_size)
      response = self._execute_with_retry(request)

  def list_topic_subscriptions(self, topic, page_size=1000):
    """List topic subscriptions."""
    request = self._api_client().projects().topics().subscriptions().list(
        topic=topic, pageSize=page_size)
    response = self._execute_with_retry(request)

    while True:
      for subscription in response.get('subscriptions', []):
        yield subscription

      next_page_token = response.get('nextPageToken')
      if not next_page_token:
        break

      request = self._api_client().projects().topics().subscriptions().list(
          topic=topic, pageToken=next_page_token, pageSize=page_size)
      response = self._execute_with_retry(request)


def project_name(project):
  return 'projects/' + project


def subscription_name(project, name):
  """Get subscription name."""
  return 'projects/{project}/subscriptions/{name}'.format(
      project=project, name=name)


def topic_name(project, name):
  """Get topic name."""
  return 'projects/{project}/topics/{name}'.format(project=project, name=name)


def parse_name(name):
  """Parse the topic or subscription name."""
  components = name.split('/')
  if len(components) != 4:
    raise ValueError('Invalid pubsub name.')

  project = components[1]
  name = components[3]
  return project, name


def _decode_data(raw_message):
  """Decode Pub/Sub data."""
  return (base64.b64decode(raw_message['data'])
          if 'data' in raw_message else None)


def raw_message_to_message(raw_message_response):
  """Convert a raw message response to a Message."""
  raw_message = raw_message_response['message']
  return Message(_decode_data(raw_message), raw_message.get('attributes'))


def _raw_message_to_received_message(client, subscription,
                                     raw_message_response):
  """Convert a raw message response to a Message."""
  raw_message = raw_message_response['message']
  return ReceivedMessage(client, subscription, _decode_data(raw_message),
                         raw_message.get('attributes'),
                         raw_message['messageId'], raw_message['publishTime'],
                         raw_message_response['ackId'])


def _message_to_dict(message):
  """Convert the message to a dict."""
  result = {}

  if message.data:
    result['data'] = base64.b64encode(message.data).decode('utf-8')

  if message.attributes:
    result['attributes'] = message.attributes

  return result
